# Timetable Widget, ZPT widget that renders complex timetables
#
# Copyright (C) 2000-2002 Juan David Ibez Palomar <jdavid@nuxeo.com>
#               2000-2001 Universitat Jaume I (http://www.uji.es)
#               2002      Nuxeo (http://www.nuxeo.com)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""Timetable Widget"""
__version__ = "$Revision: 1.1.1.1 $"

from types import StringType


def mcm(l):
    """
    Calculates the minimum common multiple of a list of integers.
    """

    if len(l) == 0:
        return 0

    l.sort()
    l.reverse()
    x = l[0]
    if x == 0:
        x = 1
    for y in l[1:]:
        if y==0:
            y = 1
        # Which is the maximum and which is hte minimum?
        maximum = max(x, y)
        minimum = min(x, y)
        # Calculate the maximum common factor (mcf)
        mod = maximum % minimum
        while mod:
            maximum = minimum
            minimum = mod
            mod = maximum % minimum
        mcf = minimum
        # Calculate the mcm
        x = mcf*(x/mcf)*(y/mcf)
    return x


class Time:
    """
    This class represents a positive amount of time with a precision of
    minutes. Internally time is represented as a pair of values, hours
    and minutes.
    """

    def __init__(self, time):
        """
        The constructor takes an string with the format "hh:mm", it
        parses the string and internally stores the hours and minutes
        as integers.
        """

        x = time.split(':')

        if len(x) != 2:
            raise ValueError, time

        self.hh, self.mm = int(x[0]), int(x[1])

    def __cmp__(self, other):
        """Compares two Time instances."""

        sh, sm = self.hh, self.mm
        oh, om = other.hh, other.mm
        if sh > oh:
            return 1
        elif sh == oh:
            if sm > om:
                return 1
            elif sm == om:
                return 0
            else:
                return -1
        else:
            return -1

    def __sub__(self, other):
        """
        Calculates the time between two Time objects, the result
        is a real number, the units is hours.
        """

        sh, sm = self.hh, self.mm
        oh, om = other.hh, other.mm
        return (sh - oh) + (sm - om)/60.0

    def __str__(self):
        """
        Textual representation of a Time object, it's "hh:mm".
        """

        return "%02d:%02d" % (self.hh, self.mm)



class Cell:
    """
    Class which represents a cell in the table. A cell can be:

      - new:

      - busy:

      - free:

    It can have the attributes:

      - content:

      - start:

      - end:

      - rowspan:

      - colspan:
    """

    new, busy, free = 0, 1, 2

    def __init__(self, type, content=None, start=None, end=None,
                 rowspan=None, colspan=None):
        self.type = type

        if content is not None:
            self.content = content

        if start is not None:
            self.start = start

        if end is not None:
            self.end = end

        if rowspan is not None:
            self.rowspan = rowspan

        if colspan is not None:
            self.colspan = colspan


class Row:
    def __init__(self):
        self.cells = []

    def __getitem__(self, i):
        return self.cells[i]

    def __setitem__(self, i, x):
        self.cells[i] = x

    def __len__(self):
        return len(self.cells)

    def append(self, x):
        self.cells.append(x)

class TimeGrid:
  def __init__(self):
    self.items = []

  def __getitem__(self, i):
    return self.items[i]

  def __len__(self):
    return len(self.items)

  def set(self, start, end, item):
    for i in range(len(self.items)):
      istart, iend, iitem = self.items[i]
      if start < istart or (start == istart and end <= iend):
        index = i
        break
    else:
      index = len(self.items)
    self.items.insert(index, (start, end, item))


  def render(self, times, mode='html'):
      """ """
      nitems, iitems = len(self.items), 0
      # blocks = [(nrows, ncols), ..]
      blocks = []
      nrows = 0
      table = []
      state = []
      for time in times[:-1]:
          row = Row()
          table.append(row)

          # add the busy cells
          for rowspan in state:
              if rowspan > 0:
                  row.append(Cell(Cell.busy))
              else:
                  row.append(Cell(Cell.free))

          # add new cells
          irow = 0
          while iitems < nitems:
            start, end, item = self.items[iitems]
            if start == time:
                # look for a free cell
                while irow < len(row):
                    if row[irow].type == Cell.free:
                        break
                    irow = irow + 1

                # add cell
                rowspan = times.index(end) - times.index(start)
                cell = Cell(Cell.new, item, start, end, rowspan)
                if irow >= len(row):
                    state.append(rowspan)
                    row.append(cell)
                else:
                    state[irow] = rowspan
                    row[irow] = cell

                # next item
                iitems = iitems + 1
            else:
                break

          ncols = len(row)
          # empty row?
          if ncols == 0:
              row.append(Cell(Cell.free))

          # next row, reduce the current rowspans
          nrows = nrows + 1
          for i in range(ncols):
              rowspan = state[i]
              if rowspan > 0:
                  state[i] = rowspan - 1

          # a new block?
          if state.count(0) == ncols:
              state = []
              blocks.append((nrows, ncols))
              nrows = 0

      # calculate the number of columns
      total_ncols = mcm(map(lambda x: x[1], blocks))

      # add colspan to each row and fill the incomplete rows with free cells
      base = 0
      for nrows, ncols in blocks:
          try:
              colspan = total_ncols/ncols
          except ZeroDivisionError:
              colspan = total_ncols
          irow = 0
          while irow < nrows:
              row = table[base + irow]
              row.colspan = colspan
              icol = len(row)
              while icol < ncols:
                  row.append(Cell(Cell.free))
                  icol = icol + 1
              irow = irow + 1
          base = base + nrows

      # If the mode is html, merge the adjacents free cells
      if mode == 'html':
          for row in table:
              i = 0
              while i<len(row.cells):
                  cell = row.cells[i]
                  if cell.type == Cell.free:
                      j = i + 1
                      while j<len(row.cells) and row.cells[j].type==Cell.free:
                          row.cells[j].type = Cell.busy
                          j = j + 1
                      if j-i > 1:
                          cell.colspan = row.colspan * (j-i)
                      i = j
                  else:
                      i = i + 1
                  

      return table, total_ncols


class TimeGridsCollection:
    """ """

    def __init__(self, time_grids, headers, times, full_day_events):
        if filter(lambda x: x, full_day_events):
            full_day = []
        else:
            full_day = None

        if headers is None:
            self.headers = None
        else:
            self.headers = []

        cols = []
        for i in range(len(time_grids)):
            table, ncols = time_grids[i].render(times)
            if headers is not None:
                self.headers.append({'header': headers[i], 'width': ncols})

            if full_day is not None:
                full_day.append({'events': full_day_events[i], 'width': ncols})

            cols.append(table)

        self.full_day_events = full_day


        self.body = []
        for i in range(len(times)-1):
            items = []
            for col in cols:
                items.append(col[i])
            self.body.append({'start': times[i], 'end': times[i+1],
                              'items': items})
